# python3
#
# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Example using TF Lite to detect objects with the Raspberry Pi camera."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import argparse
import io
import re
import time
import os

from annotation import Annotator

import cv2
import numpy as np

from PIL import Image
from tflite_runtime.interpreter import Interpreter

CAMERA_WIDTH = 640
CAMERA_HEIGHT = 480

DISP_CAMERA_WIDTH = 800
DISP_CAMERA_HEIGHT = 600
DISP_OBJLIST_WIDTH = 300
DISP_OBJLIST_HEIGHT = 600

# display: 1920x1080
OVERLAY_CAMERA_XPOS = 0
OVERLAY_CAMERA_YPOS = 0
OVERLAY_OBJLIST_XPOS = 800
OVERLAY_OBJLIST_YPOS = 0

def load_labels(path):
    """Loads the labels file. Supports files with or without index numbers."""
    with open(path, 'r', encoding='utf-8') as f:
        lines = f.readlines()
        labels = {}
        for row_number, content in enumerate(lines):
            pair = re.split(r'[:\s]+', content.strip(), maxsplit=1)
            if len(pair) == 2 and pair[0].strip().isdigit():
                labels[int(pair[0])] = pair[1].strip()
            else:
                labels[row_number] = pair[0].strip()
    return labels


def set_input_tensor(interpreter, image):
    """Sets the input tensor."""
    tensor_index = interpreter.get_input_details()[0]['index']
    input_tensor = interpreter.tensor(tensor_index)()[0]
    input_tensor[:, :] = image


def get_output_tensor(interpreter, index):
    """Returns the output tensor at the given index."""
    output_details = interpreter.get_output_details()[index]
    tensor = np.squeeze(interpreter.get_tensor(output_details['index']))
    return tensor


def detect_objects(interpreter, image, threshold):
    """Returns a list of detection results, each a dictionary of object info."""
    set_input_tensor(interpreter, image)
    interpreter.invoke()

    # Get all output details
    boxes = get_output_tensor(interpreter, 0)
    classes = get_output_tensor(interpreter, 1)
    scores = get_output_tensor(interpreter, 2)
    count = int(get_output_tensor(interpreter, 3))

    results = []
    for i in range(count):
        if scores[i] >= threshold:
            result = {
                'bounding_box': boxes[i],
                'class_id': classes[i],
                'score': scores[i]
            }
            results.append(result)
    return results


def annotate_objects(annotator, results, labels, CAMERA_WIDTH, CAMERA_HEIGHT):
    """Draws the bounding box and label for each object in the results."""
    for obj in results:
        # Convert the bounding box figures from relative coordinates
        # to absolute coordinates based on the original resolution
        ymin, xmin, ymax, xmax = obj['bounding_box']
        xmin = int(xmin * CAMERA_WIDTH)
        xmax = int(xmax * CAMERA_WIDTH)
        ymin = int(ymin * CAMERA_HEIGHT)
        ymax = int(ymax * CAMERA_HEIGHT)

        # Overlay the box, label, and score on the camera preview
        annotator.bounding_box([xmin, ymin, xmax, ymax])
        annotator.text([xmin, ymin],
                       '%s\n%.2f' % (labels[obj['class_id']], obj['score']))

class ObjList:
    def __init__(self):
        self.clear_canvas()


    def clear_canvas(self):
        self.canvas = np.zeros((DISP_OBJLIST_HEIGHT, DISP_OBJLIST_WIDTH, 3), dtype='uint8')
        self.canvas[:] = [128, 128, 128]


    def print_results(self, results, labels):
        ypos = 30
        self.clear_canvas()
        for obj in results:
            cv2.putText(self.canvas,
                        '{:15s}'.format(labels[obj['class_id']]),
                        (10, ypos),
                        cv2.FONT_HERSHEY_COMPLEX,
                        1,
                        (0, 0, 0),
                        2,
                        cv2.LINE_AA
            )
            cv2.putText(self.canvas,
                        '{:.2f}'.format(obj['score']),
                        (220, ypos),
                        cv2.FONT_HERSHEY_COMPLEX,
                        1,
                        (0, 0, 0),
                        2,
                        cv2.LINE_AA
            )
            ypos += 40

def overlay_imgs(bg, fg, xpos, ypos):
    frame = bg
    height, width = fg.shape[:2]
    frame[ypos:height + ypos, xpos:width + xpos] = fg
    return frame

def main():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        '--model', help='File path of .tflite file.', required=True)
    parser.add_argument(
        '--labels', help='File path of labels file.', required=True)
    parser.add_argument(
        '--threshold',
        help='Score threshold for detected objects.',
        required=False,
        type=float,
        default=0.4)
    parser.add_argument(
        '--camera_id',
        help='video device number N (ex. /dev/videoN)',
        required=False,
        type=int,
        default=5)
    args = parser.parse_args()

    labels = load_labels(args.labels)
    interpreter = Interpreter(args.model)
    interpreter.allocate_tensors()
    _, input_height, input_width, _ = interpreter.get_input_details()[0]['shape']

    camera = cv2.VideoCapture(args.camera_id + cv2.CAP_V4L2)

    try:
        if camera.isOpened() is False:
            raise("IO Error")
        camera.set(cv2.CAP_PROP_FRAME_WIDTH, 800)
        camera.set(cv2.CAP_PROP_FRAME_HEIGHT, 600)
        CAMERA_WIDTH = int(camera.get(cv2.CAP_PROP_FRAME_WIDTH))
        CAMERA_HEIGHT = int(camera.get(cv2.CAP_PROP_FRAME_HEIGHT))
        annotator = Annotator(camera, default_color=(0, 0, 128))
        canvas = ObjList()
        cv2.namedWindow('Object Detection Demo', cv2.WINDOW_NORMAL)
        bg = cv2.imread(os.path.join(os.path.dirname(__file__), 'bg.jpg'))
        cv2.imshow('Object Detection Demo', bg)
        while True:
            annotator.update()
            image = annotator.buffer.resize((input_width, input_height), Image.BICUBIC)
            start_time = time.monotonic()
            results = detect_objects(interpreter, image, args.threshold)
            elapsed_ms = (time.monotonic() - start_time) * 1000
            canvas.print_results(results, labels)
            annotate_objects(annotator, results, labels, CAMERA_WIDTH, CAMERA_HEIGHT)
            annotator.text([5, 0], '%.1fms' % (elapsed_ms))
            frame = cv2.cvtColor(np.asarray(annotator.buffer), cv2.COLOR_RGB2BGR
)
            frame = overlay_imgs(bg, cv2.resize(frame, dsize=(DISP_CAMERA_WIDTH,
 DISP_CAMERA_HEIGHT)), OVERLAY_CAMERA_XPOS, OVERLAY_CAMERA_YPOS)
            frame = overlay_imgs(frame, canvas.canvas, OVERLAY_OBJLIST_XPOS, OVERLAY_OBJLIST_YPOS)
            cv2.imshow('Object Detection Demo', frame)
            cv2.waitKey(1)
            del image
    except KeyboardInterrupt:
        camera.release()
        cv2.destroyAllWindows()

if __name__ == '__main__':
    main()
